package me.totalfreedom.totalfreedommod.sql;

import me.totalfreedom.totalfreedommod.FreedomService;
import me.totalfreedom.totalfreedommod.admin.Admin;
import me.totalfreedom.totalfreedommod.banning.Ban;
import me.totalfreedom.totalfreedommod.player.PlayerData;
import me.totalfreedom.totalfreedommod.util.FLog;
import me.totalfreedom.totalfreedommod.util.FUtil;

import java.sql.*;
import java.text.MessageFormat;
import java.util.UUID;

public class SQLite extends FreedomService
{
    private final String FILE_NAME = "database.db";

    private Connection connection;

    @Override
    public void onStart()
    {
        connect();
        checkTables();
    }

    @Override
    public void onStop()
    {
        disconnect();
    }

    public void connect()
    {
        try
        {
            connection = DriverManager.getConnection("jdbc:sqlite:" + plugin.getDataFolder() + "/" + FILE_NAME);
            FLog.info("Successfully connected to the database.");
        } catch (SQLException e)
        {
            FLog.severe("Failed to connect to the database: " + e.getMessage());
        }
    }

    public void disconnect()
    {
        try
        {
            if (connection != null)
            {
                connection.close();
            }
        } catch (SQLException e)
        {
            FLog.severe("Failed to disconnect from the database: " + e.getMessage());
        }
    }

    public void checkTables()
    {
        try
        {
            DatabaseMetaData meta = connection.getMetaData();
            if (tableNotExists(meta, "bans"))
            {
                createBanTable();
            }

            if (tableNotExists(meta, "admins"))
            {
                createAdminsTable();
            }
            if (tableNotExists(meta, "players"))
            {
                createPlayersTable();
            }
        } catch (SQLException e)
        {
            FLog.severe("Failed to check tables on database: " + e.getMessage());
        }
    }

    private void createPlayersTable()
    {
        try (PreparedStatement statement = connection.prepareStatement("CREATE TABLE `players` (`uuid` VARCHAR NOT NULL, `ips` VARCHAR NOT NULL, `notes` VARCHAR, `tag` VARCHAR, `discord_id` VARCHAR, `master_builder` BOOLEAN NOT NULL, `ride_mode` VARCHAR NOT NULL, `coins` INT, `items` VARCHAR, `total_votes` INT NOT NULL, `display_discord` BOOLEAN NOT NULL, `login_message` VARCHAR, `inspect` BOOLEAN NOT NULL);"))
        {
            statement.executeUpdate();
        } catch (SQLException e)
        {
            FLog.severe("Failed to create the players table: " + e.getMessage());
        }
    }

    private void createAdminsTable()
    {
        try (PreparedStatement statement = connection.prepareStatement("CREATE TABLE `admins` (`uuid` VARCHAR NOT NULL, `ips` VARCHAR NOT NULL, `rank` VARCHAR NOT NULL, `active` BOOLEAN NOT NULL, `last_login` LONG NOT NULL, `command_spy` BOOLEAN NOT NULL, `potion_spy` BOOLEAN NOT NULL, `ac_format` VARCHAR);"))
        {
            statement.executeUpdate();
        } catch (SQLException e)
        {
            FLog.severe("Failed to create the admins table: " + e.getMessage());
        }
    }

    private void createBanTable()
    {
        try (PreparedStatement statement = connection.prepareStatement("CREATE TABLE `bans` (`name` VARCHAR, `uuid` VARCHAR, `ips` VARCHAR, `by` VARCHAR NOT NULL, `at` LONG NOT NULL, `expires` LONG, `reason` VARCHAR);"))
        {
            statement.executeUpdate();
        } catch (SQLException e)
        {
            FLog.severe("Failed to create the bans table: " + e.getMessage());
        }
    }

    public void truncate(String table)
    {
        try (PreparedStatement statement = connection.prepareStatement("DELETE FROM ?"))
        {
            statement.setString(1, table);
            statement.executeUpdate();
        } catch (SQLException e)
        {
            FLog.severe("Failed to truncate " + table + ": " + e.getMessage());
        }
    }

    public ResultSetProvider getBanList() throws SQLException
    {
        try (PreparedStatement statement = connection.prepareStatement("SELECT * FROM bans"))
        {
            return ResultSetProvider.provideAllRows(statement.executeQuery()).join();
        }
    }

    public ResultSetProvider getAdminList() throws SQLException
    {
        try (PreparedStatement statement = connection.prepareStatement("SELECT * FROM bans"))
        {
            return ResultSetProvider.provideAllRows(statement.executeQuery()).join();
        }
    }

    public void setAdminValue(Admin admin, String key, Object value)
    {
        try
        {
            Object[] data = {key, admin.getUuid()};
            PreparedStatement statement = connection.prepareStatement(MessageFormat.format("UPDATE admins SET {0}=? WHERE uuid=''{1}''", data));
            statement = setUnknownType(statement, 1, value);
            statement.executeUpdate();

        } catch (SQLException e)
        {
            FLog.severe("Failed to update admin value:");
            FLog.severe(e);
        }
    }

    public void setPlayerValue(PlayerData player, String key, Object value)
    {
        try
        {
            Object[] data = {key, player.getUuid()};
            PreparedStatement statement = connection.prepareStatement(MessageFormat.format("UPDATE players SET {0}=? WHERE uuid=''{1}''", data));
            statement = setUnknownType(statement, 1, value);
            statement.executeUpdate();

        } catch (SQLException e)
        {
            FLog.severe("Failed to update player value: " + e.getMessage());
        }
    }

    public PreparedStatement setUnknownType(PreparedStatement statement, int index, Object value) throws SQLException
    {
        if (value == null)
        {
            statement.setString(index, null);
        } else if (value.getClass().equals(String.class))
        {
            String v = (String) value;
            statement.setString(index, v);
        } else if (value.getClass().equals(Integer.class))
        {
            int v = (int) value;
            statement.setInt(index, v);
        } else if (value.getClass().equals(Boolean.class))
        {
            boolean v = (boolean) value;
            statement.setBoolean(index, v);
        } else if (value.getClass().equals(Long.class))
        {
            long v = (long) value;
            statement.setLong(index, v);
        }
        return statement;
    }

    public Object getValue(ResultSetProvider provider, String key, Object value) throws SQLException
    {
        Object result = null;
        if (value instanceof String)
        {
            result = provider.getString(key);
        } else if (value instanceof Integer)
        {
            result = provider.getInt(key);
        } else if (value instanceof Boolean)
        {
            result = provider.getBoolean(key);
        } else if (value instanceof Long)
        {
            result = provider.getLong(key);
        }
        return result;
    }

    public void addAdmin(Admin admin)
    {

        try (PreparedStatement statement = connection.prepareStatement("INSERT INTO admins VALUES (?, ?, ?, ?, ?, ?, ?, ?)"))
        {
            statement.setString(1, admin.getUuid().toString());
            statement.setString(2, FUtil.listToString(admin.getIps()));
            statement.setString(3, admin.getRank().toString());
            statement.setBoolean(4, admin.isActive());
            statement.setLong(5, admin.getLastLogin().getTime());
            statement.setBoolean(6, admin.getCommandSpy());
            statement.setBoolean(7, admin.getPotionSpy());
            statement.setString(8, admin.getAcFormat());
            statement.executeUpdate();
        } catch (SQLException e)
        {
            FLog.severe("Failed to add admin:");
            FLog.severe(e);
        }
    }

    public void addPlayer(PlayerData player)
    {
        try (PreparedStatement statement = connection.prepareStatement("INSERT INTO players VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)"))
        {
            statement.setString(1, player.getUuid().toString());
            statement.setString(2, FUtil.listToString(player.getIps()));
            statement.setString(3, FUtil.listToString(player.getNotes()));
            statement.setString(4, FUtil.miniMessage(player.getTag()));
            statement.setString(5, player.getDiscordID());
            statement.setBoolean(6, player.isMasterBuilder());
            statement.setString(7, player.getRideMode().name());
            statement.setInt(8, player.getCoins());
            statement.setString(9, FUtil.listToString(player.getItems()));
            statement.setInt(10, player.getTotalVotes());
            statement.setBoolean(11, player.doesDisplayDiscord());
            statement.setString(12, player.getLoginMessage());
            statement.setBoolean(13, player.hasInspection());
            statement.executeUpdate();
        } catch (SQLException e)
        {
            FLog.severe("Failed to add player:");
            FLog.severe(e);
        }
    }

    public ResultSetProvider getAdminByUuid(UUID uuid)
    {
        try (PreparedStatement statement = connection.prepareStatement("SELECT * FROM admins WHERE uuid=?"))
        {
            statement.setString(1, uuid.toString());
            return ResultSetProvider.provideSpecificRow(statement.executeQuery()).join();
        } catch (SQLException e)
        {
            FLog.severe("Failed to get admin by name:");
            FLog.severe(e);
        }

        return null;
    }

    public ResultSetProvider getPlayerByUuid(UUID uuid)
    {
        try (PreparedStatement statement = connection.prepareStatement("SELECT * FROM players WHERE uuid=?"))
        {
            statement.setString(1, uuid.toString());
            return ResultSetProvider.provideSpecificRow(statement.executeQuery()).join();
        } catch (SQLException e)
        {
            FLog.severe("Failed to get player by UUID:");
            FLog.severe(e);
        }

        return null;
    }

    public ResultSetProvider getMasterBuilders()
    {
        try (PreparedStatement statement = connection.prepareStatement("SELECT * FROM players WHERE master_builder=true"))
        {
            return ResultSetProvider.provideAllRows(statement.executeQuery()).join();
        } catch (SQLException e)
        {
            FLog.severe("Failed to get Master Builders:");
            FLog.severe(e);
        }

        return null;
    }

    public ResultSetProvider getPlayerByIp(String ip)
    {
        try (PreparedStatement statement = connection.prepareStatement("SELECT * FROM players WHERE ips LIKE %?%"))
        {
            statement.setString(1, ip);
            return ResultSetProvider.provideSpecificRow(statement.executeQuery()).join();
        } catch (SQLException e)
        {
            FLog.severe("Failed to get player by ip:");
            FLog.severe(e);
        }

        return null;
    }

    public void removeAdmin(Admin admin)
    {
        try (PreparedStatement statement = connection.prepareStatement("DELETE FROM admins where name=?"))
        {
            statement.setString(1, admin.getName());
            statement.executeUpdate();
        } catch (SQLException e)
        {
            FLog.severe("Failed to remove admin:");
            FLog.severe(e);
        }
    }

    public void addBan(Ban ban)
    {
        try (PreparedStatement statement = connection.prepareStatement("INSERT INTO bans VALUES (?, ?, ?, ?, ?, ?, ?)"))
        {
            statement.setString(1, ban.getUsername());
            String uuid = null;
            if (ban.hasUUID())
            {
                uuid = ban.getUuid().toString();
            }
            statement.setString(2, uuid);
            statement.setString(3, FUtil.listToString(ban.getIps()));
            statement.setString(4, ban.getBy());
            statement.setLong(5, ban.getAt().getTime());
            statement.setLong(6, ban.getExpiryUnix());
            statement.setString(7, ban.getReason());
            statement.executeUpdate();
        } catch (SQLException e)
        {
            FLog.severe("Failed to add ban: " + e.getMessage());
        }
    }

    public void removeBan(Ban ban)
    {
        try
        {
            try (PreparedStatement statement = connection.prepareStatement("DELETE FROM bans WHERE name=?"))
            {
                statement.setString(1, ban.getUsername());
                statement.executeUpdate();
            }

            for (String ip : ban.getIps())
            {
                try (PreparedStatement statement = connection.prepareStatement("DELETE FROM bans WHERE ips LIKE %?%"))
                {
                    statement.setString(1, ip);
                    statement.executeUpdate();
                }
            }
        } catch (SQLException e)
        {
            FLog.severe("Failed to remove ban: " + e.getMessage());
        }
    }

    // We've changed this to read tableNotExists because it's more accurate in context.
    public boolean tableNotExists(DatabaseMetaData meta, String name) throws SQLException
    {
        return !meta.getTables(null, null, name, null).next();
    }
}